# References:
# - https://github.com/qemu/qemu/blob/master/block/qcow2.c
# - https://github.com/qemu/qemu/blob/master/docs/interop/qcow2.txt
from __future__ import annotations

import sys
import zlib
from functools import cached_property, lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import AlignedStream

from dissect.hypervisor.disk.c_qcow2 import (
    NORMAL_SUBCLUSTER_TYPES,
    QCOW2_MAGIC,
    QCOW2_MAGIC_BYTES,
    UNALLOCATED_SUBCLUSTER_TYPES,
    ZERO_SUBCLUSTER_TYPES,
    QCow2ClusterType,
    QCow2SubclusterType,
    c_qcow2,
    ctz,
)
from dissect.hypervisor.exceptions import Error, InvalidHeaderError

if TYPE_CHECKING:
    from collections.abc import Iterator

try:
    if sys.version_info >= (3, 14):
        from compression import zstd  # novermin
    else:
        from backports import zstd
    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False


class QCow2:
    """QCOW2 virtual disk implementation.

    If a data-file is required and ``fh`` is not a ``Path``, it's required to manually pass a file like object
    in the `data_file` argument. Otherwise, the data file will be automatically opened if it exists in the same directory.
    It's possible to defer opening the data file by passing ``allow_no_data_file=True``.

    The same applies to the backing-file. This too can be deferred by passing ``allow_no_backing_file=True``.

    Args:
        fh: File handle or path to the QCOW2 file.
        data_file: Optional file handle for the data file. If not provided and ``fh`` is a ``Path``, it will try to open it automatically.
        backing_file: Optional file handle for the backing file. If not provided and ``fh`` is a ``Path``, it will try to open it automatically.
        allow_no_data_file: If True, allows the QCOW2 file to be opened without a data file.
        allow_no_backing_file: If True, allows the QCOW2 file to be opened without a backing file.
    """  # noqa: E501

    def __init__(
        self,
        fh: BinaryIO | Path,
        data_file: BinaryIO | None = None,
        backing_file: BinaryIO | None = None,
        *,
        allow_no_data_file: bool = False,
        allow_no_backing_file: bool = False,
    ):
        if isinstance(fh, Path):
            self.path = fh
            self.fh = self.path.open("rb")
        else:
            self.path = None
            self.fh = fh

        self.fh.seek(0)

        self.header = c_qcow2.QCowHeader(self.fh)
        if self.header.magic != QCOW2_MAGIC:
            raise InvalidHeaderError("Invalid qcow2 header magic")

        if self.header.version < 2 or self.header.version > 3:
            raise InvalidHeaderError(f"Unsupported qcow2 version: {self.header.version}")

        if self.header.cluster_bits < c_qcow2.MIN_CLUSTER_BITS or self.header.cluster_bits > c_qcow2.MAX_CLUSTER_BITS:
            raise InvalidHeaderError(f"Unsupported cluster size: 2**{self.header.cluster_bits}")

        self.size = self.header.size

        self.cluster_bits = self.header.cluster_bits
        self.cluster_size = 1 << self.cluster_bits
        self.subclusters_per_cluster = c_qcow2.QCOW_EXTL2_SUBCLUSTERS_PER_CLUSTER if self.has_subclusters else 1
        self.subcluster_size = self.cluster_size // self.subclusters_per_cluster
        self.subcluster_bits = ctz(self.subcluster_size)

        self._l2_entry_size = c_qcow2.L2E_SIZE_EXTENDED if self.has_subclusters else c_qcow2.L2E_SIZE_NORMAL
        self.l2_bits = self.cluster_bits - ctz(self._l2_entry_size)
        self.l2_size = 1 << self.l2_bits

        # 104 = byte offset of compression_type
        if self.header.header_length > 104:
            self.compression_type = self.header.compression_type
        else:
            self.compression_type = c_qcow2.QCOW2_COMPRESSION_TYPE_ZLIB

        if self.compression_type == c_qcow2.QCOW2_COMPRESSION_TYPE_ZSTD and not HAS_ZSTD:
            raise RuntimeError("zstandard module not available")

        self.csize_shift = 62 - (self.cluster_bits - 8)
        self.csize_mask = (1 << (self.cluster_bits - 8)) - 1
        self.cluster_offset_mask = (1 << self.csize_shift) - 1

        if self.subcluster_size < (1 << c_qcow2.MIN_CLUSTER_BITS):
            raise InvalidHeaderError(f"Unsupported subcluster size: 0x{self.subcluster_size:x}")

        if self.header.crypt_method:
            raise NotImplementedError("Encrypted qcow2 files are not supported")

        self.backing_format = None
        self.feature_table = None
        self.crypto_header = None
        self.bitmap_header = None
        self.image_data_file = None
        self.unknown_extensions = []
        self._read_extensions()

        if self.needs_data_file:
            self.data_file = self._open_data_file(data_file, allow_no_data_file)
        else:
            self.data_file = self.fh

        self.backing_file = None
        self.auto_backing_file = None
        self.image_backing_file = None
        if self.header.backing_file_offset:
            self.fh.seek(self.header.backing_file_offset)
            self.auto_backing_file = self.fh.read(self.header.backing_file_size).decode()
            self.image_backing_file = self.auto_backing_file.upper()

            self.backing_file = self._open_backing_file(backing_file, allow_no_backing_file)

        self.l2_table = lru_cache(128)(self.l2_table)

    def _read_extensions(self) -> None:
        start_offset = self.header.header_length
        end_offset = self.header.backing_file_offset or 1 << self.cluster_bits

        offset = start_offset
        while offset < end_offset:
            self.fh.seek(offset)
            ext = c_qcow2.QCowExtension(self.fh)
            offset += len(c_qcow2.QCowExtension)

            if offset > end_offset or ext.len > end_offset - offset:
                break  # Invalid extension, just ignore

            if ext.magic == c_qcow2.QCOW2_EXT_MAGIC_END:
                break

            if ext.magic == c_qcow2.QCOW2_EXT_MAGIC_BACKING_FORMAT:
                self.backing_format = self.fh.read(ext.len).decode().upper()
                self.image_backing_format = self.backing_format.upper()
            elif ext.magic == c_qcow2.QCOW2_EXT_MAGIC_FEATURE_TABLE:
                self.feature_table = self.fh.read(ext.len)
            elif ext.magic == c_qcow2.QCOW2_EXT_MAGIC_CRYPTO_HEADER:
                self.crypto_header = c_qcow2.Qcow2CryptoHeaderExtension(self.fh)
            elif ext.magic == c_qcow2.QCOW2_EXT_MAGIC_BITMAPS:
                self.bitmap_header = c_qcow2.Qcow2BitmapHeaderExt(self.fh)
            elif ext.magic == c_qcow2.QCOW2_EXT_MAGIC_DATA_FILE:
                self.image_data_file = self.fh.read(ext.len).decode()
            else:
                self.unknown_extensions.append((ext, self.fh.read(ext.len)))

            # Align to nearest 8 byte boundary
            offset += (ext.len + 7) & 0xFFFFFFF8

    def _open_data_file(self, data_file: BinaryIO | None, allow_no_data_file: bool = False) -> BinaryIO | None:
        if data_file is not None:
            return data_file

        if self.path:
            if (data_file_path := self.path.parent.joinpath(self.image_data_file)).exists():
                return data_file_path.open("rb")

            if not allow_no_data_file:
                raise Error(f"data-file {str(data_file_path)!r} not found (image_data_file = {self.image_data_file!r})")
        elif allow_no_data_file:
            return None

        raise Error(f"data-file required but not provided (image_data_file = {self.image_data_file!r})")

    def _open_backing_file(self, backing_file: BinaryIO | None, allow_no_backing_file: bool = False) -> BinaryIO | None:
        backing_file_path = None
        if backing_file is None:
            if self.path:
                if (backing_file_path := self.path.parent.joinpath(self.auto_backing_file)).exists():
                    backing_file = backing_file_path.open("rb")
                elif not allow_no_backing_file:
                    raise Error(
                        f"backing-file {str(backing_file_path)!r} not found (auto_backing_file = {self.auto_backing_file!r})"  # noqa: E501
                    )
            elif not allow_no_backing_file:
                raise Error(f"backing-file required but not provided (auto_backing_file = {self.auto_backing_file!r})")

        if backing_file:
            if backing_file.read(4) == QCOW2_MAGIC_BYTES:
                if backing_file_path:
                    backing_file.close()
                    backing_file = QCow2(backing_file_path).open()
                else:
                    backing_file = QCow2(backing_file).open()
            else:
                backing_file.seek(0)

        return backing_file

    @cached_property
    def snapshots(self) -> list[QCow2Snapshot]:
        snapshots: list[QCow2Snapshot] = []

        offset = self.header.snapshots_offset
        for _ in range(self.header.nb_snapshots):
            snapshots.append(QCow2Snapshot(self, offset))
            offset += (snapshots[-1].entry_size + 7) & ~7  # Round up to 8 bytes

        return snapshots

    @cached_property
    def l1_table(self) -> list[int]:
        # L1 table is usually relatively small, it can be at most 32MB on PB or EB size disks
        self.fh.seek(self.header.l1_table_offset)
        return c_qcow2.uint64[self.header.l1_size](self.fh)

    def l2_table(self, l2_offset: int) -> L2Table:
        return L2Table(self, l2_offset)

    @property
    def has_backing_file(self) -> bool:
        return self.backing_file is not None

    @property
    def needs_backing_file(self) -> bool:
        return self.header.backing_file_offset != 0

    @property
    def has_data_file(self) -> bool:
        return self.data_file is not None and self.data_file != self.fh

    @property
    def needs_data_file(self) -> bool:
        return bool(self.header.incompatible_features & c_qcow2.QCOW2_INCOMPAT_DATA_FILE)

    @property
    def has_subclusters(self) -> bool:
        return bool(self.header.incompatible_features & c_qcow2.QCOW2_INCOMPAT_EXTL2)

    def open(self) -> QCow2Stream:
        """Open the QCow2 file for reading."""
        if self.needs_data_file and not self.has_data_file:
            raise Error(f"data-file required but not provided (image_data_file = {self.image_data_file!r})")
        if self.needs_backing_file and not self.has_backing_file:
            raise Error(f"backing-file required but not provided (auto_backing_file = {self.auto_backing_file!r})")
        return QCow2Stream(self)


class QCow2Snapshot:
    """Wrapper class for snapshot table entries."""

    def __init__(self, qcow2: QCow2, offset: int):
        self.qcow2 = qcow2
        self.offset = offset

        self.qcow2.fh.seek(offset)
        self.header = c_qcow2.QCowSnapshotHeader(self.qcow2.fh)

        # Older versions may not have all the extra data fields
        # Instead of reading them manually, just pad the extra data to fit our struct
        extra_data = self.qcow2.fh.read(self.header.extra_data_size)
        self.extra = c_qcow2.QCowSnapshotExtraData(extra_data.ljust(len(c_qcow2.QCowSnapshotExtraData), b"\x00"))

        unknown_extra_size = self.header.extra_data_size - len(c_qcow2.QCowSnapshotExtraData)
        self.unknown_extra = self.qcow2.fh.read(unknown_extra_size) if unknown_extra_size > 0 else None

        self.id = self.qcow2.fh.read(self.header.id_str_size).decode()
        self.name = self.qcow2.fh.read(self.header.name_size).decode()

        self.entry_size = self.qcow2.fh.tell() - offset

    def open(self) -> QCow2Stream:
        """Open the snapshot for reading."""
        return QCow2Stream(self.qcow2, self.l1_table)

    @cached_property
    def l1_table(self) -> list[int]:
        # L1 table is usually relatively small, it can be at most 32MB on PB or EB size disks
        self.qcow2.fh.seek(self.header.l1_table_offset)
        return c_qcow2.uint64[self.header.l1_size](self.qcow2.fh)


class QCow2Stream(AlignedStream):
    """Aligned stream for reading QCow2 files."""

    def __init__(self, qcow2: QCow2, l1_table: list[int] | None = None):
        super().__init__(qcow2.header.size, align=qcow2.cluster_size)
        self.qcow2 = qcow2
        self.l1_table = l1_table or qcow2.l1_table

    def _yield_runs(self, offset: int, length: int) -> Iterator[tuple[QCow2SubclusterType, int, int, int]]:
        # reference: qcow2_get_host_offset
        while length > 0:
            sc_type = None
            host_offset = 0
            read_count = 0

            l1_index = offset_to_l1_index(self.qcow2, offset)
            l2_index = offset_to_l2_index(self.qcow2, offset)
            sc_index = offset_to_sc_index(self.qcow2, offset)

            offset_in_cluster = offset_into_cluster(self.qcow2, offset)

            bytes_needed = length + offset_in_cluster
            # at the time being we just use the entire l2 table and not cached slices
            # this is actually the bytes available/remaining in this l2 table
            bytes_available = (self.qcow2.l2_size - l2_index) << self.qcow2.cluster_bits
            bytes_needed = min(bytes_needed, bytes_available)

            if l1_index > self.qcow2.header.l1_size:
                # bytes_needed is already the smaller value here
                read_count = bytes_needed - offset_in_cluster

                yield (QCow2SubclusterType.QCOW2_SUBCLUSTER_UNALLOCATED_PLAIN, offset, host_offset, read_count)

                length -= read_count
                offset += read_count
                continue

            l2_offset = self.l1_table[l1_index] & c_qcow2.L1E_OFFSET_MASK
            if not l2_offset:
                # bytes_needed is already the smaller value here
                read_count = bytes_needed - offset_in_cluster

                yield (QCow2SubclusterType.QCOW2_SUBCLUSTER_UNALLOCATED_PLAIN, offset, host_offset, read_count)

                length -= read_count
                offset += read_count
                continue

            l2_table = self.qcow2.l2_table(l2_offset)
            l2_entry = l2_table.entry(l2_index)
            l2_bitmap = l2_table.bitmap(l2_index)

            sc_type = get_subcluster_type(self.qcow2, l2_entry, l2_bitmap, sc_index)

            if sc_type == QCow2SubclusterType.QCOW2_SUBCLUSTER_COMPRESSED:
                host_offset = l2_entry & c_qcow2.L2E_COMPRESSED_OFFSET_SIZE_MASK
            elif sc_type in NORMAL_SUBCLUSTER_TYPES:
                host_cluster_offset = l2_entry & c_qcow2.L2E_OFFSET_MASK
                host_offset = host_cluster_offset + offset_in_cluster

            nb_clusters = size_to_clusters(self.qcow2, bytes_needed)
            sc_count = count_contiguous_subclusters(self.qcow2, nb_clusters, sc_index, l2_table, l2_index)
            # this is the amount of contiguous bytes available of the same subcluster type
            bytes_available = (sc_count + sc_index) << self.qcow2.subcluster_bits

            # account for the offset in the cluster
            read_count = min(bytes_available, bytes_needed) - offset_in_cluster

            yield (sc_type, offset, host_offset, read_count)
            length -= read_count
            offset += read_count

    def _read_compressed(self, cluster_descriptor: int, offset: int, length: int) -> bytes:
        offset_in_cluster = offset_into_cluster(self.qcow2, offset)
        coffset = cluster_descriptor & self.qcow2.cluster_offset_mask
        nb_csectors = ((cluster_descriptor >> self.qcow2.csize_shift) & self.qcow2.csize_mask) + 1
        # Original source uses the mask ~(~(QCOW2_COMPRESSED_SECTOR_SIZE - 1ULL))
        # However bit inversion is weird in Python, and this evaluates to 511, so we use that value instead.
        csize = nb_csectors * c_qcow2.QCOW2_COMPRESSED_SECTOR_SIZE - (coffset & 511)

        self.qcow2.fh.seek(coffset)
        buf = self.qcow2.fh.read(csize)
        decompressed = self._decompress(buf)

        return decompressed[offset_in_cluster : offset_in_cluster + length]

    def _decompress(self, buf: bytes) -> bytes:
        if self.qcow2.compression_type == c_qcow2.QCOW2_COMPRESSION_TYPE_ZLIB:
            dctx = zlib.decompressobj(-12)
            return dctx.decompress(buf, self.qcow2.cluster_size)

        if self.qcow2.compression_type == c_qcow2.QCOW2_COMPRESSION_TYPE_ZSTD:
            dctx = zstd.ZstdDecompressor()
            return dctx.decompress(buf, self.qcow2.cluster_size)

        raise Error(f"Invalid compression type: {self.qcow2.compression_type}")

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        for sc_type, read_offset, run_offset, run_length in self._yield_runs(offset, length):
            unalloc_zeroed = sc_type in UNALLOCATED_SUBCLUSTER_TYPES and not self.qcow2.has_backing_file

            if sc_type in ZERO_SUBCLUSTER_TYPES or unalloc_zeroed:
                result.append(b"\x00" * run_length)
            elif sc_type in UNALLOCATED_SUBCLUSTER_TYPES and self.qcow2.has_backing_file:
                self.qcow2.backing_file.seek(read_offset)
                result.append(self.qcow2.backing_file.read(run_length))
            elif sc_type == QCow2SubclusterType.QCOW2_SUBCLUSTER_COMPRESSED:
                result.append(self._read_compressed(run_offset, read_offset, run_length))
            elif sc_type == QCow2SubclusterType.QCOW2_SUBCLUSTER_NORMAL:
                self.qcow2.data_file.seek(run_offset)
                result.append(self.qcow2.data_file.read(run_length))

        return b"".join(result)


class L2Table:
    """Convenience class for accessing the L2 table."""

    def __init__(self, qcow2: QCow2, offset: int):
        self.qcow2 = qcow2
        self.offset = offset

        l2_table_size = self.qcow2.l2_size * (self.qcow2._l2_entry_size // 8)
        self.qcow2.fh.seek(offset)
        self._table = c_qcow2.uint64[l2_table_size](self.qcow2.fh)

    def entry(self, idx: int) -> int:
        return self._table[idx * self.qcow2._l2_entry_size // 8]

    def bitmap(self, idx: int) -> int:
        if self.qcow2.has_subclusters:
            return self._table[(idx * self.qcow2._l2_entry_size // 8) + 1]
        return 0


def offset_into_cluster(qcow2: QCow2, offset: int) -> int:
    return offset & (qcow2.cluster_size - 1)


def offset_into_subcluster(qcow2: QCow2, offset: int) -> int:
    return offset & (qcow2.subcluster_size - 1)


def size_to_clusters(qcow2: QCow2, size: int) -> int:
    return (size + (qcow2.cluster_size - 1)) >> qcow2.cluster_bits


def size_to_subclusters(qcow2: QCow2, size: int) -> int:
    return (size + (qcow2.subcluster_size - 1)) >> qcow2.subcluster_bits


def offset_to_l1_index(qcow2: QCow2, offset: int) -> int:
    return offset >> (qcow2.l2_bits + qcow2.cluster_bits)


def offset_to_l2_index(qcow2: QCow2, offset: int) -> int:
    return (offset >> qcow2.cluster_bits) & (qcow2.l2_size - 1)


def offset_to_sc_index(qcow2: QCow2, offset: int) -> int:
    return (offset >> qcow2.subcluster_bits) & (qcow2.subclusters_per_cluster - 1)


def get_cluster_type(qcow2: QCow2, l2_entry: int) -> QCow2ClusterType:
    if l2_entry & c_qcow2.QCOW_OFLAG_COMPRESSED:
        return QCow2ClusterType.QCOW2_CLUSTER_COMPRESSED

    if (l2_entry & c_qcow2.QCOW_OFLAG_ZERO) and not qcow2.has_subclusters:
        if l2_entry & c_qcow2.L2E_OFFSET_MASK:
            return QCow2ClusterType.QCOW2_CLUSTER_ZERO_ALLOC
        return QCow2ClusterType.QCOW2_CLUSTER_ZERO_PLAIN

    if not l2_entry & c_qcow2.L2E_OFFSET_MASK:
        if qcow2.has_data_file and l2_entry & c_qcow2.QCOW_OFLAG_COPIED:
            return QCow2ClusterType.QCOW2_CLUSTER_NORMAL
        return QCow2ClusterType.QCOW2_CLUSTER_UNALLOCATED

    return QCow2ClusterType.QCOW2_CLUSTER_NORMAL


def get_subcluster_type(qcow2: QCow2, l2_entry: int, l2_bitmap: int, sc_index: int) -> QCow2SubclusterType:
    c_type = get_cluster_type(qcow2, l2_entry)

    sc_alloc_mask = 1 << sc_index
    sc_zero_mask = sc_alloc_mask << 32

    if qcow2.has_subclusters:
        if c_type == QCow2ClusterType.QCOW2_CLUSTER_COMPRESSED:
            return QCow2SubclusterType.QCOW2_SUBCLUSTER_COMPRESSED
        if c_type == QCow2ClusterType.QCOW2_CLUSTER_NORMAL:
            if (l2_bitmap >> 32) & l2_bitmap:
                return QCow2SubclusterType.QCOW2_SUBCLUSTER_INVALID
            if l2_bitmap & sc_zero_mask:  # QCOW_OFLAG_SUB_ZERO(sc_index)
                return QCow2SubclusterType.QCOW2_SUBCLUSTER_ZERO_ALLOC
            if l2_bitmap & sc_alloc_mask:  # QCOW_OFLAG_SUB_ALLOC(sc_index)
                return QCow2SubclusterType.QCOW2_SUBCLUSTER_NORMAL
            return QCow2SubclusterType.QCOW2_SUBCLUSTER_UNALLOCATED_ALLOC
        if c_type == QCow2ClusterType.QCOW2_CLUSTER_UNALLOCATED:
            if l2_bitmap & ((1 << 32) - 1):
                return QCow2SubclusterType.QCOW2_SUBCLUSTER_INVALID
            if l2_bitmap & sc_zero_mask:  # QCOW_OFLAG_SUB_ZERO(sc_index)
                return QCow2SubclusterType.QCOW2_SUBCLUSTER_ZERO_PLAIN
            return QCow2SubclusterType.QCOW2_SUBCLUSTER_UNALLOCATED_PLAIN

        raise Error(f"Invalid cluster type: {c_type}")

    if c_type == QCow2ClusterType.QCOW2_CLUSTER_COMPRESSED:
        return QCow2SubclusterType.QCOW2_SUBCLUSTER_COMPRESSED
    if c_type == QCow2ClusterType.QCOW2_CLUSTER_ZERO_PLAIN:
        return QCow2SubclusterType.QCOW2_SUBCLUSTER_ZERO_PLAIN
    if c_type == QCow2ClusterType.QCOW2_CLUSTER_ZERO_ALLOC:
        return QCow2SubclusterType.QCOW2_SUBCLUSTER_ZERO_ALLOC
    if c_type == QCow2ClusterType.QCOW2_CLUSTER_NORMAL:
        return QCow2SubclusterType.QCOW2_SUBCLUSTER_NORMAL
    if c_type == QCow2ClusterType.QCOW2_CLUSTER_UNALLOCATED:
        return QCow2SubclusterType.QCOW2_SUBCLUSTER_UNALLOCATED_PLAIN

    raise Error(f"Invalid cluster type: {c_type}")


def get_subcluster_range_type(
    qcow2: QCow2, l2_entry: int, l2_bitmap: int, sc_from: int
) -> tuple[QCow2SubclusterType, int]:
    sc_type = get_subcluster_type(qcow2, l2_entry, l2_bitmap, sc_from)

    # No subclusters, so count the entire cluster
    if not qcow2.has_subclusters or sc_type == QCow2SubclusterType.QCOW2_SUBCLUSTER_COMPRESSED:
        return sc_type, qcow2.subclusters_per_cluster - sc_from

    sc_mask = (1 << sc_from) - 1
    if sc_type == QCow2SubclusterType.QCOW2_SUBCLUSTER_NORMAL:
        val = l2_bitmap | sc_mask  # QCOW_OFLAG_SUB_ALLOC_RANGE(0, sc_from)
        return sc_type, ctz(val) - sc_from
    if sc_type in ZERO_SUBCLUSTER_TYPES:
        val = (l2_bitmap | sc_mask) >> 32  # QCOW_OFLAG_SUB_ZERO_RANGE(0, sc_from)
        return sc_type, ctz(val) - sc_from
    if sc_type in UNALLOCATED_SUBCLUSTER_TYPES:
        # We need to mask it with a 64bit mask because Python flips the sign bit
        inv_mask = ~sc_mask & ((1 << 64) - 1)  # ~QCOW_OFLAG_SUB_ALLOC_RANGE(0, sc_from)

        val = ((l2_bitmap >> 32) | l2_bitmap) & inv_mask
        return sc_type, ctz(val) - sc_from

    raise Error(f"Invalid subcluster type: {sc_type}")


def count_contiguous_subclusters(
    qcow2: QCow2, nb_clusters: int, sc_index: int, l2_table: L2Table, l2_index: int
) -> int:
    count = 0
    expected_type = None
    expected_offset = None

    check_offset = False
    check_offset_types = (
        QCow2SubclusterType.QCOW2_SUBCLUSTER_NORMAL,
        QCow2SubclusterType.QCOW2_SUBCLUSTER_ZERO_ALLOC,
        QCow2SubclusterType.QCOW2_SUBCLUSTER_UNALLOCATED_ALLOC,
    )

    for i in range(nb_clusters):
        first_sc = sc_index if i == 0 else 0
        l2_entry = l2_table.entry(l2_index + i)
        l2_bitmap = l2_table.entry(l2_index + i)

        sc_type, sc_count = get_subcluster_range_type(qcow2, l2_entry, l2_bitmap, first_sc)

        if i == 0:
            if sc_type == QCow2SubclusterType.QCOW2_SUBCLUSTER_COMPRESSED:
                return sc_count

            expected_type = sc_type
            expected_offset = l2_entry & c_qcow2.L2E_OFFSET_MASK
            check_offset = sc_type in check_offset_types
        elif sc_type != expected_type:
            break
        elif check_offset:
            expected_offset += qcow2.cluster_size
            if expected_offset != l2_entry & c_qcow2.L2E_OFFSET_MASK:
                break

        count += sc_count
        if first_sc + sc_count < qcow2.subclusters_per_cluster:
            break

    return count



from __future__ import annotations

from bisect import bisect_right
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from defusedxml import ElementTree
from dissect.util.stream import AlignedStream

from dissect.hypervisor.disk.c_hdd import SECTOR_SIZE, c_hdd
from dissect.hypervisor.exceptions import InvalidHeaderError

if TYPE_CHECKING:
    from collections.abc import Iterator
    from xml.etree.ElementTree import Element

DEFAULT_TOP_GUID = UUID("{5fbaabe3-6958-40ff-92a7-860e329aab41}")
NULL_GUID = UUID("00000000-0000-0000-0000-000000000000")


class HDD:
    """Parallels HDD virtual disk implementation.

    Args:
        path: The path to the .hdd directory or .hdd file in a .hdd directory.
    """

    def __init__(self, path: Path):
        if path.is_file():
            path = path.parent
        self.path = path

        descriptor_path = path.joinpath("DiskDescriptor.xml")
        if not descriptor_path.exists():
            raise ValueError(f"Invalid Parallels HDD path: {path} (missing DiskDescriptor.xml)")

        self.descriptor = Descriptor(descriptor_path)

    def _open_image(self, path: Path) -> BinaryIO:
        """Helper method for opening image files relative to this HDD.

        Args:
            path: The path to the image file to open.
        """
        root = self.path
        filename = path.name

        if path.is_absolute():
            # If the path is absolute, check if it exists
            if not path.exists():
                # If the absolute path does not exist, we're probably dealing with a HDD
                # that's been copied or moved (e.g., uploaded or copied as evidence)
                # Try a couple of common patterns to see if we can locate the file
                #
                # Example variables:
                # root = /some/path/example.pvm/example.hdd/
                # path = /other/path/absolute.pvm/absolute.hdd/absolute.ext

                # File is in same HDD directory
                # candidate_path = /some/path/example.pvm/example.hdd/absolute.ext
                candidate_path = root / filename
                if not candidate_path.exists():
                    # File is in a separate HDD directory in parent (VM) directory
                    # candidate_path = /some/path/example.pvm/absolute.hdd/absolute.ext
                    candidate_path = root.parent / path.parent.name / filename

                if not candidate_path.exists():
                    # File is in .pvm directory in parent of parent directory (linked clones)
                    # candidate_path = /some/path/absolute.pvm/absolute.hdd/absolute.ext
                    candidate_path = root.parent.parent / path.parent.parent.name / path.parent.name / filename

                path = candidate_path

            return path.open("rb")

        # If the path is relative, it's always relative to the HDD root
        return (root / path).open("rb")

    def open(self, guid: str | UUID | None = None) -> BinaryIO:
        """Open a stream for this HDD, optionally for a specific snapshot.

        If no snapshot GUID is provided, the "top" snapshot will be used.

        Args:
            guid: The snapshot GUID to open.
        """
        if guid and not isinstance(guid, UUID):
            guid = UUID(guid)

        if guid is None:
            guid = self.descriptor.snapshots.top_guid or DEFAULT_TOP_GUID

        chain = self.descriptor.get_snapshot_chain(guid)

        streams = []
        for storage in self.descriptor.storage_data.storages:
            stream = None
            for guid in chain[::-1]:
                image = storage.find_image(guid)
                fh = self._open_image(Path(image.file))

                if image.type == "Compressed":
                    fh = HDS(fh, parent=stream)
                elif image.type != "Plain":
                    raise ValueError(f"Unsupported image type: {image.type}")

                stream = fh

            streams.append((storage, stream))

        return StorageStream(streams)


class Descriptor:
    """Helper class for working with ``DiskDescriptor.xml``.

    References:
        - https://github.com/qemu/qemu/blob/master/docs/interop/prl-xml.txt

    Args:
        path: The path to ``DiskDescriptor.xml``.
    """

    def __init__(self, path: Path):
        self.path = path

        self.xml: Element = ElementTree.fromstring(path.read_text())
        self.storage_data = StorageData.from_xml(self.xml.find("StorageData"))
        self.snapshots = Snapshots.from_xml(self.xml.find("Snapshots"))

    def get_snapshot_chain(self, guid: UUID) -> list[UUID]:
        """Return the snapshot chain for a given snapshot GUID.

        Args:
            guid: The snapshot GUID to return a chain for.
        """
        shot = self.snapshots.find_shot(guid)

        chain = [shot.guid]
        while shot.parent != NULL_GUID:
            shot = self.snapshots.find_shot(shot.parent)
            chain.append(shot.guid)

        return chain


@dataclass
class XMLEntry:
    @classmethod
    def from_xml(cls, element: Element) -> XMLEntry:
        if element.tag != cls.__name__:
            raise ValueError(f"Invalid {cls.__name__} XML element")
        return cls._from_xml(element)

    @classmethod
    def _from_xml(cls, element: Element) -> XMLEntry:
        raise NotImplementedError


@dataclass
class StorageData(XMLEntry):
    storages: list[Storage]

    @classmethod
    def _from_xml(cls, element: Element) -> StorageData:
        return cls(list(map(Storage.from_xml, element.iterfind("Storage"))))


@dataclass
class Storage(XMLEntry):
    start: int
    end: int
    images: list[Image]

    @classmethod
    def _from_xml(cls, element: Element) -> Storage:
        start = int(element.find("Start").text)
        end = int(element.find("End").text)
        images = list(map(Image.from_xml, element.iterfind("Image")))

        return cls(start, end, images)

    def find_image(self, guid: UUID) -> Image:
        """Find a specific image GUID.

        Args:
            guid: The image GUID to find.

        Raises:
            KeyError: If the GUID could not be found.
        """
        for image in self.images:
            if image.guid == guid:
                return image

        raise KeyError(f"Image GUID not found: {guid}")


@dataclass
class Image(XMLEntry):
    guid: UUID
    type: str
    file: str

    @classmethod
    def _from_xml(cls, element: Element) -> Image:
        return cls(
            UUID(element.find("GUID").text),
            element.find("Type").text,
            element.find("File").text,
        )


@dataclass
class Snapshots(XMLEntry):
    top_guid: UUID | None
    shots: list[Shot]

    @classmethod
    def _from_xml(cls, element: Element) -> Snapshots:
        top_guid = element.find("TopGUID")
        if top_guid:
            top_guid = UUID(top_guid.text)
        shots = list(map(Shot.from_xml, element.iterfind("Shot")))

        return cls(top_guid, shots)

    def find_shot(self, guid: UUID) -> Shot:
        """Find a specific snapshot GUID.

        Args:
            guid: The snapshot GUID to find.

        Raises:
            KeyError: If the GUID could not be found.
        """
        for shot in self.shots:
            if shot.guid == guid:
                return shot

        raise KeyError(f"Shot GUID not found: {guid}")


@dataclass
class Shot(XMLEntry):
    guid: UUID
    parent: UUID

    @classmethod
    def _from_xml(cls, element: Element) -> Shot:
        return cls(
            UUID(element.find("GUID").text),
            UUID(element.find("ParentGUID").text),
        )


class StorageStream(AlignedStream):
    """Stream implementation for HDD streams.

    HDD files can exist of one or multiple streams, starting at consecutive offsets.
    This class stitches all streams together into a single stream.

    Args:
        streams: A list of :class:`Storage` and file-like object tuples.
    """

    def __init__(self, streams: list[tuple[Storage, BinaryIO]]):
        self.streams = sorted(streams, key=lambda entry: entry[0].start)
        self._lookup = []

        size = 0
        for storage, _ in self.streams:
            self._lookup.append(storage.start)
            size = storage.end

        super().__init__(size * SECTOR_SIZE)

    def _read(self, offset: int, length: int) -> bytes:
        sector = offset // SECTOR_SIZE
        count = (length + SECTOR_SIZE - 1) // SECTOR_SIZE

        result = []
        stream_idx = bisect_right(self._lookup, sector) - 1

        while count > 0 and stream_idx < len(self.streams):
            storage, stream = self.streams[stream_idx]
            sectors_remaining = storage.end - sector
            read_sectors = min(sectors_remaining, count)

            stream.seek((sector - storage.start) * SECTOR_SIZE)
            result.append(stream.read(read_sectors * SECTOR_SIZE))

            sector += read_sectors
            count -= read_sectors
            stream_idx += 1

        return b"".join(result)


class HDS(AlignedStream):
    """Parallels HDS implementation.

    HDS is the format for Parallels sparse disk files.

    Args:
        fh: The file-like object to the HDS file.
        parent: Optional file-like object for the parent HDS file.
    """

    def __init__(self, fh: BinaryIO, parent: BinaryIO | None = None):
        self.fh = fh
        self.parent = parent

        self.header = c_hdd.pvd_header(fh)
        if self.header.m_Sig not in (c_hdd.SIGNATURE_STRUCTURED_DISK_V1, c_hdd.SIGNATURE_STRUCTURED_DISK_V2):
            raise InvalidHeaderError(f"Invalid HDS header signature: {self.header.m_Sig}")

        if self.header.m_Sig == c_hdd.SIGNATURE_STRUCTURED_DISK_V1:
            size = self.header.m_SizeInSectors_v1
            self._bat_step = self.header.m_Sectors
            self._bat_multiplier = 1
        else:
            size = self.header.m_SizeInSectors_v2
            self._bat_step = 1
            self._bat_multiplier = self.header.m_Sectors

        self.cluster_size = self.header.m_Sectors * SECTOR_SIZE

        self.data_offset = self.header.m_FirstBlockOffset
        self.in_use = self.header.m_DiskInUse == c_hdd.SIGNATURE_DISK_IN_USE

        super().__init__(size * SECTOR_SIZE)

    @cached_property
    def bat(self) -> list[int]:
        """Return the block allocation table (BAT)."""
        self.fh.seek(len(c_hdd.pvd_header))
        return c_hdd.uint32[self.header.m_Size](self.fh)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        for read_offset, read_size in self._iter_runs(offset, length):
            # Sentinel value for sparse runs
            if read_offset is None:
                if self.parent:
                    self.parent.seek(offset)
                    result.append(self.parent.read(read_size))
                else:
                    result.append(b"\x00" * read_size)
            else:
                self.fh.seek(read_offset)
                result.append(self.fh.read(read_size))

            offset += read_size
            length -= read_size

        return b"".join(result)

    def _iter_runs(self, offset: int, length: int) -> Iterator[tuple[int, int]]:
        """Iterate optimized read runs for a given offset and read length.

        Args:
            offset: The offset in bytes to generate runs for.
            length: The length in bytes to generate runs for.
        """
        bat = self.bat

        run_offset = None
        run_size = 0

        while offset < self.size and length > 0:
            cluster_idx, offset_in_cluster = divmod(offset, self.cluster_size)
            read_size = min(self.cluster_size - offset_in_cluster, length)

            bat_entry = bat[cluster_idx]
            # BAT entry of 0 means either a sparse or a parent read
            # Use 0 to denote a sparse run for now to make calculations easier
            read_offset = 0 if bat_entry == 0 else bat_entry * self._bat_multiplier * SECTOR_SIZE + offset_in_cluster

            if run_offset is None:
                # First iteration
                run_offset = read_offset
                run_size = read_size
            elif (read_offset == run_offset + run_size) or (run_offset, read_offset) == (0, 0):
                # Consecutive (sparse) clusters
                run_size += read_size
            else:
                # New run
                # Replace 0 with None as sparse sentinel value
                yield (run_offset or None, run_size)

                # Reset run
                run_offset = read_offset
                run_size = read_size

            offset += read_size
            length -= read_size

        if run_offset is not None:
            # Flush remaining run
            # Replace 0 with None as sparse sentinel value
            yield (run_offset or None, run_size)



from __future__ import annotations

from dissect.cstruct import cstruct

hyperv_def = """
/* ======== File header ======== */

#define SIGNATURE_STORAGE_HEADER            0x01282014
#define FIRST_HEADER_OFFSET                 0x00000000
#define SECOND_HEADER_OFFSET                0x00001000

struct HyperVStorageHeader {
    uint32      signature;                  // 0x01282014
    uint32      checksum;                   // Checksum of this header (0x2e), with this field set to 0
    uint16      sequence_number;            // Header sequence number
    uint32      version;                    // Format version
    uint64      unknown2;                   //
    uint32      alignment;                  // Can't be over 0x10000
    uint64      replay_log_offset;          // Offset of the HyperVStorageReplayLog header
    uint64      replay_log_size;            // Log size
    uint32      header_size;                // File header size
};

/* ======== Replay log ======== */

#define SIGNATURE_REPLAY_LOG_HEADER         0x01110003

struct HyperVStorageReplayLog {
    uint32      signature;                  // 0x01110003
    uint32      checksum;                   // Checksum of this header (0x22), with this field set to 0
    uint32      num_entries;
    uint8       unknown1;
    uint32      max_entries;
    uint32      unknown2;
    uint32      unknown3;
    uint32      unknown4;
    uint32      unknown5;
    uint8       unknown6;
};

struct HyperVStorageReplayLogEntry {
    uint64      offset;
    uint32      size;
    uint32      unknown2;
    uint32      unknown3;
    uint32      checksum;                   // Checksum of this entry (0x1C), with this field set to 0
    uint32      data_checksum;              // Checksum of the data for this log entry
};

/* ======== Object tables ======== */

#define SIGNATURE_OBJECT_TABLE_HEADER       0x01110001
#define OBJECT_TABLE_OFFSET                 0x00002000

enum ObjectEntryType : uint8 {
    Unknown0 = 0,                           // <unknown(0)>
    ObjectTable = 1,                        // Object Table
    KeyTable = 2,                           // Key File
    File = 3,                               // File
    Free = 4,                               // Free
    Unknown5Header = 5,                     // <unknown(5)>Header
    ReplayLog = 6,                          // Replay Log
    ChangeTrackingBuffer = 7,               // Change Tracking Buffer
};

struct HyperVStorageObjectTable {
    uint32      signature;                  // 0x01110001
    uint32      num_entries;
};

struct HyperVStorageObjectTableEntry {
    ObjectEntryType type;                   // Entry Type
    uint32      checksum;                   // Checksum of this entry (0x12), with this field set to 0
    uint64      offset;                     // Entry offset
    uint32      size;                       // Entry size
    uint8       allocated;                  // Allocated flag
};

/* ======== Key tables ======== */

#define SIGNATURE_KEY_TABLE_HEADER          0x0002

enum KeyDataType : uint8 {
    Free = 1,
    Unknown = 2,
    Int = 3,
    UInt = 4,
    Double = 5,
    String = 6,
    Array = 7,
    Bool = 8,
    Node = 9,
};

flag KeyDataFlag : uint8 {
    FileObjectPointer = 0x01,
};

struct HyperVStorageKeyTable {
    uint16      signature;                  // 0x0002
    uint16      index;                      // Table index
    uint16      sequence_number;            // Table sequence number
    uint32      checksum;                   // Checksum of this header (0xA), with this field set to 0
};

struct HyperVStorageKeyTableEntryHeader {
    uint16      type;                       // Must be <= KeyDataType.Node
    uint32      size;                       // Size of this entry
    uint16      parent_table_idx;           // Table index of the parent node
    uint32      parent_offset;              // Offset of the parent node in the parent table
    uint32      checksum;                   // Reference: HyperVStorageKeyTableEntry::CalculateChecksum
    uint32      insertion_sequence;         // Reference: HyperVStorageKeyTableEntry::SetInsertionSequence
    uint8       data_offset;                // Offset into the entry data to get to the value data
};
"""

c_hyperv = cstruct().load(hyperv_def)

ObjectEntryType = c_hyperv.ObjectEntryType
KeyDataType = c_hyperv.KeyDataType
KeyDataFlag = c_hyperv.KeyDataFlag



from __future__ import annotations

import struct

from dissect.cstruct import cstruct

vmdk_def = """
typedef struct {
    char    magic[4];                           // Magic "KDMV" LE
    uint32  version;                            // Version
    uint32  flags;                              // Flags
    uint64  capacity;                           // The maximum data number of sectors (capacity)
    uint64  grain_size;                         // The grain number of sectors
    uint64  descriptor_offset;                  // The descriptor sector number
    uint64  descriptor_size;                    // The descriptor number of sectors
    uint32  num_grain_table_entries;            // The number of grain table entries
    uint64  secondary_grain_directory_offset;   // The secondary grain directory sector number
    uint64  primary_grain_directory_offset;     // The primary grain directory sector number
    uint64  overhead;                           // The metadata (overhead) number of sectors
    uint8   is_dirty;                           // Value to indicate the VMDK was cleanly closed
    char    single_end_line_char;               // The single end of line character
    char    non_end_line_char;                  // A non end of line character
    char    double_end_line_chars[2];           // The double end of line characters
    uint16  compress_algorithm;                 // The compression method
    char    pad[433];                           // Padding
} VMDKSparseExtentHeader;

typedef struct {
    char    magic[4];                           // Magic "COWD" LE
    uint32  version;                            // Version
    uint32  flags;                              // Flags
    uint32  capacity;                           // The maximum data number of sectors (capacity)
    uint32  grain_size;                         // The grain number of sectors
    uint32  primary_grain_directory_offset;     // The primary grain directory sector number
    uint32  num_grain_directory_entries;        // The number of grain table entries
    uint32  next_free_grain;                    // The next free grain

    //uint32  num_cylinders;                      // The number of cylinders
    //uint32  num_heads;                          // The number of heads
    //uint32  num_sectors;                        // The number of sectors

    //char    parent_filename[1024];              // The parent filename
    //uint32  parent_generation;                  // The parent generation

    //uint32  generation;                         // The generation
    //char    name[60];                           // The name
    //char    description[512];                   // The description
    //uint32  saved_generation;                   // The saved generation
    //uint64  reserved;                           // Reserved
    //uint8   is_dirty;                           // Value to indicate the COWD was cleanly closed
    //char    padding[396];                       // Padding
} COWDSparseExtentHeader;

typedef struct {
    uint64  magic;
    uint64  version;
    uint64  capacity;
    uint64  grain_size;
    uint64  grain_table_size;
    uint64  flags;
    uint64  reserved1;
    uint64  reserved2;
    uint64  reserved3;
    uint64  reserved4;
    uint64  volatile_header_offset;
    uint64  volatile_header_size;
    uint64  journal_header_offset;
    uint64  journal_header_size;
    uint64  journal_offset;
    uint64  journal_size;
    uint64  grain_directory_offset;
    uint64  grain_directory_size;
    uint64  grain_tables_offset;
    uint64  grain_tables_size;
    uint64  free_bitmap_offset;
    uint64  free_bitmap_size;
    uint64  backmap_offset;
    uint64  backmap_size;
    uint64  grains_offset;
    uint64  grains_size;
    uint8   pad[304];
} VMDKSESparseConstHeader;

typedef struct {
    uint64  magic;
    uint64  free_gt_number;
    uint64  next_txn_seq_number;
    uint64  replay_journal;
    uint8   pad[480];
} VMDKSESparseVolatileHeader;

#define SPARSE_MAGICNUMBER                  0x564D444B
#define SPARSE_VERSION_INCOMPAT_FLAGS       3
#define SPARSE_GTE_EMPTY                    0x00000000
#define SPARSE_GD_AT_END                    0xFFFFFFFFFFFFFFFF
#define SPARSE_SINGLE_END_LINE_CHAR         '\n'
#define SPARSE_NON_END_LINE_CHAR            ' '
#define SPARSE_DOUBLE_END_LINE_CHAR1        '\r'
#define SPARSE_DOUBLE_END_LINE_CHAR2        '\n'
#define SPARSEFLAG_COMPAT_FLAGS             0x0000FFFF
#define SPARSEFLAG_VALID_NEWLINE_DETECTOR   1
#define SPARSEFLAG_USE_REDUNDANT            2
#define SPARSEFLAG_MAGIC_GTE                4
#define SPARSEFLAG_INCOMPAT_FLAGS           0xFFFF0000
#define SPARSEFLAG_COMPRESSED               0x10000
#define SPARSEFLAG_EMBEDDED_LBA             0x20000
#define SPARSE_COMPRESSALGORITHM_NONE       0x0000
#define SPARSE_COMPRESSALGORITHM_DEFLATE    0x0001

#define SESPARSE_CONST_HEADER_MAGIC         0x00000000CAFEBABE
#define SESPARSE_VOLATILE_HEADER_MAGIC      0x00000000CAFEBABE

#define SESPARSE_GRAIN_TYPE_MASK            0xF000000000000000
#define SESPARSE_GRAIN_TYPE_UNALLOCATED     0x0000000000000000
#define SESPARSE_GRAIN_TYPE_FALLTHROUGH     0x1000000000000000
#define SESPARSE_GRAIN_TYPE_ZERO            0x2000000000000000
#define SESPARSE_GRAIN_TYPE_ALLOCATED       0x3000000000000000

typedef struct {
    uint64  lba;
    uint32  cmp_size;
} SparseGrainLBAHeaderOnDisk;

typedef struct {
    uint64  lba;
    uint32  cmp_size;
    uint32  type;
} SparseSpecialLBAHeaderOnDisk;

#define GRAIN_MARKER_EOS                0
#define GRAIN_MARKER_GRAIN_TABLE        1
#define GRAIN_MARKER_GRAIN_DIRECTORY    2
#define GRAIN_MARKER_FOOTER             3
#define GRAIN_MARKER_PROGRESS           4
"""

c_vmdk = cstruct().load(vmdk_def)

SECTOR_SIZE = 512

COWD_MAGIC = b"COWD"
VMDK_MAGIC = b"KDMV"
# Technically a 8 byte header, but it's little endian so everything after the first 4 bytes is 0
SESPARSE_MAGIC = struct.pack("<I", c_vmdk.SESPARSE_CONST_HEADER_MAGIC)



from __future__ import annotations

from dissect.cstruct import cstruct

# https://github.com/VirtualBox/virtualbox/blob/main/src/VBox/Storage/VDICore.h
vdi_def = """
enum VDI_IMAGE_TYPE {
    /** Normal dynamically growing base image file. */
    NORMAL          = 1,
    /** Preallocated base image file of a fixed size. */
    FIXED,
    /** Dynamically growing image file for undo/commit changes support. */
    UNDO,
    /** Dynamically growing image file for differencing support. */
    DIFF,
};

flag VDI_IMAGE_FLAGS {
    /** Fill new blocks with zeroes while expanding image file. Only valid
     * for newly created images, never set for opened existing images. */
    ZERO_EXPAND     = 0x0100,
};

typedef struct VDIDISKGEOMETRY {
    /** Cylinders. */
    uint32_t        Cylinders;
    /** Heads. */
    uint32_t        Heads;
    /** Sectors per track. */
    uint32_t        Sectors;
    /** Sector size. (bytes per sector) */
    uint32_t        Sector;
} VDIDISKGEOMETRY, *PVDIDISKGEOMETRY;

typedef struct VDIPREHEADER {
    /** Just text info about image type, for eyes only. */
    char            szFileInfo[64];
    /** The image signature (VDI_IMAGE_SIGNATURE). */
    uint32_t        u32Signature;
    /** The image version (VDI_IMAGE_VERSION). */
    uint32_t        u32Version;
} VDIPREHEADER, *PVDIPREHEADER;

/**
 * Size of Comment field of HDD image header.
 */
#define VDI_IMAGE_COMMENT_SIZE      256

/* NOTE: All the header versions are additive, so just use the latest one. */
typedef struct VDIHEADER1PLUS {
    /** Size of this structure in bytes. */
    uint32_t        cbHeader;
    /** The image type (VDI_IMAGE_TYPE_*). */
    VDI_IMAGE_TYPE  u32Type;
    /** Image flags (VDI_IMAGE_FLAGS_*). */
    VDI_IMAGE_FLAGS fFlags;
    /** Image comment. (UTF-8) */
    char            szComment[VDI_IMAGE_COMMENT_SIZE];
    /** Offset of blocks array from the beginning of image file.
     * Should be sector-aligned for HDD access optimization. */
    uint32_t        offBlocks;
    /** Offset of image data from the beginning of image file.
     * Should be sector-aligned for HDD access optimization. */
    uint32_t        offData;
    /** Legacy image geometry (previous code stored PCHS there). */
    VDIDISKGEOMETRY LegacyGeometry;
    /** Was BIOS HDD translation mode, now unused. */
    uint32_t        u32Dummy;
    /** Size of disk (in bytes). */
    uint64_t        cbDisk;
    /** Block size. (For instance VDI_IMAGE_BLOCK_SIZE.) Should be a power of 2! */
    uint32_t        cbBlock;
    /** Size of additional service information of every data block.
     * Prepended before block data. May be 0.
     * Should be a power of 2 and sector-aligned for optimization reasons. */
    uint32_t        cbBlockExtra;
    /** Number of blocks. */
    uint32_t        cBlocks;
    /** Number of allocated blocks. */
    uint32_t        cBlocksAllocated;
    /** UUID of image. */
    char            uuidCreate[16];
    /** UUID of image's last modification. */
    char            uuidModify[16];
    /** Only for secondary images - UUID of previous image. */
    char            uuidLinkage[16];
    /** Only for secondary images - UUID of previous image's last modification. */
    char            uuidParentModify[16];
    /** LCHS image geometry (new field in VDI1.2 version. */
    VDIDISKGEOMETRY Geometry;
} VDIHEADER1PLUS, *PVDIHEADER1PLUS;
"""

c_vdi = cstruct().load(vdi_def)

VDI_IMAGE_SIGNATURE = 0xBEDA107F
VDI_IMAGE_BLOCK_FREE = ~0
VDI_IMAGE_BLOCK_ZERO = ~1



from __future__ import annotations

from dissect.cstruct import cstruct

vhd_def = """
struct footer {
    char            cookie[8];
    uint32          features;
    uint32          version;
    uint64          data_offset;
    uint32          timestamp;
    uint32          creator_application;
    uint32          creator_version;
    uint32          creator_host_os;
    uint64          original_size;
    uint64          current_size;
    uint32          disk_geometry;
    uint32          disk_type;
    uint32          checksum;
    char            unique_id[16];
    char            saved_state;
    char            reserved[426];          // Actually 427, but old versions can have a 511 byte footer
};

struct parent_locator {
    uint32          platform_code;
    uint32          platform_data_space;
    uint32          platform_data_length;
    uint32          reserved;
    uint64          platform_data_offset;
};

struct dynamic_header {
    char            cookie[8];
    uint64          data_offset;
    uint64          table_offset;
    uint32          header_version;
    uint32          max_table_entries;
    uint32          block_size;
    uint32          checksum;
    char            parent_unique_id[16];
    uint32          parent_timestamp;
    uint32          reserved1;
    char            parent_unicode_name[512];
    parent_locator  parent_locators[8];
    char            reserved2[256];
};
"""

c_vhd = cstruct(endian=">").load(vhd_def)

SECTOR_SIZE = 512



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_qcow2(__cs__.cstruct):
    MIN_CLUSTER_BITS: Literal[9] = ...
    MAX_CLUSTER_BITS: Literal[21] = ...
    QCOW2_COMPRESSED_SECTOR_SIZE: Literal[512] = ...
    QCOW2_COMPRESSION_TYPE_ZLIB: Literal[0] = ...
    QCOW2_COMPRESSION_TYPE_ZSTD: Literal[1] = ...
    L1E_SIZE: Literal[8] = ...
    L2E_SIZE_NORMAL: Literal[8] = ...
    L2E_SIZE_EXTENDED: Literal[16] = ...
    L1E_OFFSET_MASK: Literal[72057594037927424] = ...
    L2E_OFFSET_MASK: Literal[72057594037927424] = ...
    L2E_COMPRESSED_OFFSET_SIZE_MASK: Literal[4611686018427387903] = ...
    QCOW_OFLAG_COPIED: Literal[9223372036854775808] = ...
    QCOW_OFLAG_COMPRESSED: Literal[4611686018427387904] = ...
    QCOW_OFLAG_ZERO: Literal[1] = ...
    QCOW_EXTL2_SUBCLUSTERS_PER_CLUSTER: Literal[32] = ...
    QCOW2_INCOMPAT_DIRTY_BITNR: Literal[0] = ...
    QCOW2_INCOMPAT_CORRUPT_BITNR: Literal[1] = ...
    QCOW2_INCOMPAT_DATA_FILE_BITNR: Literal[2] = ...
    QCOW2_INCOMPAT_COMPRESSION_BITNR: Literal[3] = ...
    QCOW2_INCOMPAT_EXTL2_BITNR: Literal[4] = ...
    QCOW2_INCOMPAT_DIRTY: Literal[1] = ...
    QCOW2_INCOMPAT_CORRUPT: Literal[2] = ...
    QCOW2_INCOMPAT_DATA_FILE: Literal[4] = ...
    QCOW2_INCOMPAT_COMPRESSION: Literal[8] = ...
    QCOW2_INCOMPAT_EXTL2: Literal[16] = ...
    QCOW2_EXT_MAGIC_END: Literal[0] = ...
    QCOW2_EXT_MAGIC_BACKING_FORMAT: Literal[3799591626] = ...
    QCOW2_EXT_MAGIC_FEATURE_TABLE: Literal[1745090647] = ...
    QCOW2_EXT_MAGIC_CRYPTO_HEADER: Literal[87539319] = ...
    QCOW2_EXT_MAGIC_BITMAPS: Literal[595929205] = ...
    QCOW2_EXT_MAGIC_DATA_FILE: Literal[1145132097] = ...
    class QCowHeader(__cs__.Structure):
        magic: _c_qcow2.uint32
        version: _c_qcow2.uint32
        backing_file_offset: _c_qcow2.uint64
        backing_file_size: _c_qcow2.uint32
        cluster_bits: _c_qcow2.uint32
        size: _c_qcow2.uint64
        crypt_method: _c_qcow2.uint32
        l1_size: _c_qcow2.uint32
        l1_table_offset: _c_qcow2.uint64
        refcount_table_offset: _c_qcow2.uint64
        refcount_table_clusters: _c_qcow2.uint32
        nb_snapshots: _c_qcow2.uint32
        snapshots_offset: _c_qcow2.uint64
        incompatible_features: _c_qcow2.uint64
        compatible_features: _c_qcow2.uint64
        autoclear_features: _c_qcow2.uint64
        refcount_order: _c_qcow2.uint32
        header_length: _c_qcow2.uint32
        compression_type: _c_qcow2.uint8
        padding: __cs__.Array[_c_qcow2.uint8]
        @overload
        def __init__(
            self,
            magic: _c_qcow2.uint32 | None = ...,
            version: _c_qcow2.uint32 | None = ...,
            backing_file_offset: _c_qcow2.uint64 | None = ...,
            backing_file_size: _c_qcow2.uint32 | None = ...,
            cluster_bits: _c_qcow2.uint32 | None = ...,
            size: _c_qcow2.uint64 | None = ...,
            crypt_method: _c_qcow2.uint32 | None = ...,
            l1_size: _c_qcow2.uint32 | None = ...,
            l1_table_offset: _c_qcow2.uint64 | None = ...,
            refcount_table_offset: _c_qcow2.uint64 | None = ...,
            refcount_table_clusters: _c_qcow2.uint32 | None = ...,
            nb_snapshots: _c_qcow2.uint32 | None = ...,
            snapshots_offset: _c_qcow2.uint64 | None = ...,
            incompatible_features: _c_qcow2.uint64 | None = ...,
            compatible_features: _c_qcow2.uint64 | None = ...,
            autoclear_features: _c_qcow2.uint64 | None = ...,
            refcount_order: _c_qcow2.uint32 | None = ...,
            header_length: _c_qcow2.uint32 | None = ...,
            compression_type: _c_qcow2.uint8 | None = ...,
            padding: __cs__.Array[_c_qcow2.uint8] | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class QCowExtension(__cs__.Structure):
        magic: _c_qcow2.uint32
        len: _c_qcow2.uint32
        @overload
        def __init__(self, magic: _c_qcow2.uint32 | None = ..., len: _c_qcow2.uint32 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Qcow2CryptoHeaderExtension(__cs__.Structure):
        offset: _c_qcow2.uint64
        length: _c_qcow2.uint64
        @overload
        def __init__(self, offset: _c_qcow2.uint64 | None = ..., length: _c_qcow2.uint64 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Qcow2BitmapHeaderExt(__cs__.Structure):
        nb_bitmaps: _c_qcow2.uint32
        reserved32: _c_qcow2.uint32
        bitmap_directory_size: _c_qcow2.uint64
        bitmap_directory_offset: _c_qcow2.uint64
        @overload
        def __init__(
            self,
            nb_bitmaps: _c_qcow2.uint32 | None = ...,
            reserved32: _c_qcow2.uint32 | None = ...,
            bitmap_directory_size: _c_qcow2.uint64 | None = ...,
            bitmap_directory_offset: _c_qcow2.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class QCowSnapshotHeader(__cs__.Structure):
        l1_table_offset: _c_qcow2.uint64
        l1_size: _c_qcow2.uint32
        id_str_size: _c_qcow2.uint16
        name_size: _c_qcow2.uint16
        date_sec: _c_qcow2.uint32
        date_nsec: _c_qcow2.uint32
        vm_clock_nsec: _c_qcow2.uint64
        vm_state_size: _c_qcow2.uint32
        extra_data_size: _c_qcow2.uint32
        @overload
        def __init__(
            self,
            l1_table_offset: _c_qcow2.uint64 | None = ...,
            l1_size: _c_qcow2.uint32 | None = ...,
            id_str_size: _c_qcow2.uint16 | None = ...,
            name_size: _c_qcow2.uint16 | None = ...,
            date_sec: _c_qcow2.uint32 | None = ...,
            date_nsec: _c_qcow2.uint32 | None = ...,
            vm_clock_nsec: _c_qcow2.uint64 | None = ...,
            vm_state_size: _c_qcow2.uint32 | None = ...,
            extra_data_size: _c_qcow2.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class QCowSnapshotExtraData(__cs__.Structure):
        vm_state_size_large: _c_qcow2.uint64
        disk_size: _c_qcow2.uint64
        icount: _c_qcow2.uint64
        @overload
        def __init__(
            self,
            vm_state_size_large: _c_qcow2.uint64 | None = ...,
            disk_size: _c_qcow2.uint64 | None = ...,
            icount: _c_qcow2.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class QCow2ClusterType(__cs__.Enum):
        QCOW2_CLUSTER_UNALLOCATED = ...
        QCOW2_CLUSTER_ZERO_PLAIN = ...
        QCOW2_CLUSTER_ZERO_ALLOC = ...
        QCOW2_CLUSTER_NORMAL = ...
        QCOW2_CLUSTER_COMPRESSED = ...

    class QCow2SubclusterType(__cs__.Enum):
        QCOW2_SUBCLUSTER_UNALLOCATED_PLAIN = ...
        QCOW2_SUBCLUSTER_UNALLOCATED_ALLOC = ...
        QCOW2_SUBCLUSTER_ZERO_PLAIN = ...
        QCOW2_SUBCLUSTER_ZERO_ALLOC = ...
        QCOW2_SUBCLUSTER_NORMAL = ...
        QCOW2_SUBCLUSTER_COMPRESSED = ...
        QCOW2_SUBCLUSTER_INVALID = ...

# Technically `c_qcow2` is an instance of `_c_qcow2`, but then we can't use it in type hints
c_qcow2: TypeAlias = _c_qcow2

QCOW2_MAGIC: int
QCOW2_MAGIC_BYTES: bytes

QCOW2_INCOMPAT_MASK: int

QCow2ClusterType: TypeAlias = _c_qcow2.QCow2ClusterType
QCow2SubclusterType: TypeAlias = _c_qcow2.QCow2SubclusterType

NORMAL_SUBCLUSTER_TYPES: tuple[QCow2SubclusterType, ...]
ZERO_SUBCLUSTER_TYPES: tuple[QCow2SubclusterType, ...]
UNALLOCATED_SUBCLUSTER_TYPES: tuple[QCow2SubclusterType, ...]

def ctz(value: int) -> int: ...



# References:
# - https://src.openvz.org/projects/OVZ/repos/ploop/browse/include/ploop1_image.h
# - https://github.com/qemu/qemu/blob/master/docs/interop/parallels.txt
from __future__ import annotations

from dissect.cstruct import cstruct

hdd_def = """
/* Compressed disk v1 signature */
#define SIGNATURE_STRUCTURED_DISK_V1    b"WithoutFreeSpace"

/* Compressed disk v2 signature */
#define SIGNATURE_STRUCTURED_DISK_V2    b"WithouFreSpacExt"

/* Sign that the disk is in "using" state */
#define SIGNATURE_DISK_IN_USE           0x746F6E59

#define SECTOR_LOG                      9
#define SECTOR_SIZE                     (1 << SECTOR_LOG)

struct pvd_header {
    char    m_Sig[16];                              /* Signature */
    uint32  m_Type;                                 /* Disk type */
    uint32  m_Heads;                                /* heads count */
    uint32  m_Cylinders;                            /* tracks count */
    uint32  m_Sectors;                              /* Sectors per track count */
    uint32  m_Size;                                 /* Size of disk in tracks */
    union {                                         /* Size of disk in 512-byte sectors */
        struct {
            uint32  m_SizeInSectors_v1;
            uint32  Unused;
        };
        uint64  m_SizeInSectors_v2;
    };
    uint32  m_DiskInUse;                            /* Disk in use */
    uint32  m_FirstBlockOffset;                     /* First data block offset (in sectors) */
    uint32  m_Flags;                                /* Misc flags */
    uint64  m_FormatExtensionOffset;                /* Optional header offset in bytes */
};

struct pvd_ext_block_check {
    // Format Extension magic = 0xAB234CEF23DCEA87
    uint64  m_Magic;
    // Md5 checksum of the whole (without top 24 bytes of block check)
    // Format Extension Block.
    uint8   m_Md5[16];
};

struct pvd_ext_block_element_header {
    uint64  magic;
    uint64  flags;
    uint32  size;
    uint32  unused32;
};

struct pvd_dirty_bitmap_raw {
    uint64  m_Size;
    uint8   m_Id[16];
    uint32  m_Granularity;
    uint32  m_L1Size;
    uint64  m_L1[m_L1Size];
};
"""

c_hdd = cstruct().load(hdd_def)

SECTOR_SIZE = c_hdd.SECTOR_SIZE
